#include <jvmpi.h>
#include <stdlib.h>
//
#ifdef WIN32
	#include <windows.h>
	#include <hash_map>
#endif
//
#ifdef LINUX
	#include <sys/time.h>
	#include <unistd.h>
	#include <pthread.h>
#endif
//
#include <functional>
#include <algorithm>
#include <string>
#include <vector>
#include <list>
#include <queue>
#include <deque>
#include <map>
#include <set>
using namespace std;
//
#include <boost/smart_ptr.hpp>
using namespace boost;
	
	static JVMPI_Interface *jvmpi_interface = NULL;
#define CALL(f) (jvmpi_interface->f)
	
#ifdef WIN32
jlong
getPerfCounter() {
	static LONGLONG freq = 0;
	if (freq == 0) {
		LARGE_INTEGER freq2;
		QueryPerformanceFrequency( &freq2 );
		freq = freq2.QuadPart;
	}
	//
	LARGE_INTEGER counter;
	QueryPerformanceCounter( &counter );
	return counter.QuadPart;
}

jlong
getTime() {
	static LONGLONG freq = 0;
	{
		LARGE_INTEGER freq2;
		QueryPerformanceFrequency( &freq2 );
		freq = freq2.QuadPart;
	}
	//
	LARGE_INTEGER counter;
	QueryPerformanceCounter( &counter );
	return (1000 * counter.QuadPart) / freq;
}

jlong
getThreadTimeLong(HANDLE h) {
	FILETIME c, e, u, k;
	BOOL b = GetThreadTimes(h, &c, &e, &k, &u);
	LARGE_INTEGER t1;
	t1.LowPart = k.dwLowDateTime;
	t1.HighPart = k.dwHighDateTime;
	LARGE_INTEGER t2;
	t2.LowPart = u.dwLowDateTime;
	t2.HighPart = u.dwHighDateTime;
	//
	return (t1.QuadPart + t2.QuadPart) / 10000;
}

HANDLE
getRealThreadHandle(HANDLE h) {
    HANDLE hRealHandle;
    BOOL bResult = DuplicateHandle(GetCurrentProcess(),
		h,
		GetCurrentProcess(),
		&hRealHandle,
		0,
		FALSE,
		DUPLICATE_SAME_ACCESS);
    return hRealHandle;
}

class TCriticalSection 
{
	CRITICAL_SECTION m_sect;
	bool			 m_bAcquired;
public:
	TCriticalSection() { m_bAcquired = false; ::InitializeCriticalSection(&m_sect); }
	virtual ~TCriticalSection() { /*::DeleteCriticalSection(&m_sect);*/ }
	
public:
	operator CRITICAL_SECTION*() { return (CRITICAL_SECTION*) &m_sect; }
	bool IsLocked() { return m_bAcquired; }
	void Unlock() { ::LeaveCriticalSection(&m_sect); m_bAcquired = false;}
	void Lock()   { ::EnterCriticalSection(&m_sect); m_bAcquired = true;}
};
#endif

#ifdef LINUX
int GetTickCount() {
	struct timeval tv;
    gettimeofday(&tv, NULL);
    return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

class TMutex {
private:
	pthread_mutex_t m_Handle;
public:
	TMutex() {
		pthread_mutex_init(&m_Handle, NULL);
	}
	~TMutex() {
		pthread_mutex_destroy(&m_Handle);
	}
	bool Lock() {
		pthread_mutex_lock(&m_Handle);
		return true;
    }
	bool Lock(long msec) {
		const long step = 50;
		for (long value = 0; value < msec; value += step) {
			if ( pthread_mutex_trylock(&m_Handle) == 0 )
				return true;
			usleep( 1000 * step );
		}
		return false;
	}
	void Unlock() {
		pthread_mutex_unlock(&m_Handle);
	}
};
typedef TMutex TCriticalSection;
#endif


template <class T> class CLocker 
{
	T &m_o;
public:
	CLocker(T& o):m_o(o) { m_o.Lock(); }
    ~CLocker()			 { m_o.Unlock(); }
};

typedef CLocker<TCriticalSection> TCSLocker;


template< typename Out >
void split( const std :: string &s,
		   const std :: string &sep,
		   Out dest, int count = INT_MAX)
{
	int counter = 0;
	std :: string :: size_type left = s.find_first_not_of( sep );
	std :: string :: size_type right = s.find_first_of( sep, left );
	//
	while ( left < right ) {
		*dest = s.substr( left, right - left );
		++dest;
		left = s.find_first_not_of( sep, right );
		right = s.find_first_of( sep, left );
		
		if(counter++ >= count){
			*dest = s.substr( left );
			return;
		}
	}
}


class TJVMMethodInfo {
public:
	jmethodID m_ID;
	string m_Name;
	string m_Signature;
	jint m_LineStart;
	jint m_LineEnd;
	TJVMMethodInfo() {
	}
	TJVMMethodInfo(JVMPI_Method *method) {
		m_ID = method->method_id;
		m_Name = "???";
		if (method->method_name != NULL)
			m_Name = method->method_name;
		m_Signature = method->method_signature;
		m_LineStart = method->start_lineno;
		m_LineEnd = method->end_lineno;
	}
};

template<class T> struct hash2 {
	size_t operator()(const T __x) const { return (size_t)__x; }
};

class TJVMClassInfo;
#ifdef WIN32
	typedef hash_map<jmethodID, shared_ptr<TJVMMethodInfo>, hash2<jmethodID> > TIdToMethodMap;
	typedef hash_map<jmethodID, shared_ptr<TJVMClassInfo>, hash2<jmethodID> > TMethodIdToClassMap;
#endif
#ifdef LINUX
	typedef map<jmethodID, shared_ptr<TJVMMethodInfo> > TIdToMethodMap;
	typedef map<jmethodID, shared_ptr<TJVMClassInfo> > TMethodIdToClassMap;
#endif

class TJVMClassInfo {
public:
	jobjectID m_ID;
	string m_File;
	string m_Name;
	vector< shared_ptr<TJVMMethodInfo> > m_Methods;
	TIdToMethodMap m_IdToMethodMap;
	TJVMClassInfo(jobjectID id) {
		m_ID = id;
		m_Name = "unknown";
	}
	TJVMClassInfo() {
		m_Name = "unknown";
	}
	TJVMClassInfo(JVMPI_Event *event) {
		m_ID = event->u.class_load.class_id;
		if ( event->u.class_load.source_name != 0 )
			m_File = event->u.class_load.source_name;
		m_Name = "???";
		if ( event->u.class_load.class_name != 0 )
			m_Name = event->u.class_load.class_name;
		//
		for (int i = 0; i < event->u.class_load.num_methods; i++) {
			JVMPI_Method *method = event->u.class_load.methods + i;
			shared_ptr<TJVMMethodInfo> info( new TJVMMethodInfo(method) );
			m_Methods.push_back( info );
			m_IdToMethodMap[method->method_id] = info;
		}
	}
	shared_ptr<TJVMMethodInfo> getMethod(jmethodID methodID) {
		return m_IdToMethodMap[methodID];
	}
	void addMethods(TIdToMethodMap &map, TMethodIdToClassMap &map2, const shared_ptr<TJVMClassInfo> &current) {
		for (unsigned i = 0; i < m_Methods.size(); i++) {
			shared_ptr<TJVMMethodInfo> info = m_Methods[i];
			map[info->m_ID] = info;
			map2[info->m_ID] = current;
		}
	}
};

class CompareClassByID : public greater< shared_ptr<TJVMClassInfo> > {
public:
	CompareClassByID() {}
	bool operator()(const shared_ptr<TJVMClassInfo>& c1, const shared_ptr<TJVMClassInfo>& c2) const {
		return c1->m_ID < c2->m_ID;
	}
};

TCriticalSection g_SamplingCS;
TCriticalSection g_OutputCriticalSection;
string g_OutputString;
vector<string> g_Filters;
void
doOut(const char *msg, ...) {
	char buf[1024 * 16];
	va_list ap;
	va_start(ap, msg);
	vsprintf(buf, msg, ap);
	va_end(ap);
	//
	/*printf(buf);
	fflush(stdout);*/
	{
		//TCSLocker CSL( g_OutputCriticalSection );
		g_OutputString += buf;
	}
}


string getMethodName(jmethodID methodID);

bool
isVisibleMethod(jmethodID methodID) {
	string name = getMethodName(methodID);
	for (unsigned i = 0; i < g_Filters.size(); i++) {
		string filter = g_Filters[i];
		if ( filter.find("__P__") == 0 )
			if ( name.find(filter.substr(5)) != 0 )
				return false;
		if ( filter.find("__M__") == 0 )
			if ( name.find(filter.substr(5)) == 0 )
				return false;
	}
	return true;
	//return name.find("de.") == 0;
	//return true;
}

typedef map<int,int> TLineToCountMap;
typedef TLineToCountMap::iterator ILineToCountMap;

class TMethodBillInfo {
public:
	jlong m_Count;
	jlong m_DirectCount;
	jlong m_TotalTime;
	jlong m_RealTime;
    TLineToCountMap m_LineToCountMap;
	TMethodBillInfo() {
		clear();
	}
	void clear() {
		m_Count = 0;
		m_DirectCount = 0;
		m_TotalTime = 0;
		m_RealTime = 0;
		m_LineToCountMap.clear();
	}
	void update(TMethodBillInfo &bill) {
		m_Count += bill.m_Count;
		m_DirectCount += bill.m_DirectCount;
		m_TotalTime += bill.m_TotalTime;
		m_RealTime += bill.m_RealTime;
        TLineToCountMap &map = bill.m_LineToCountMap;
		for (ILineToCountMap I = map.begin(); I != map.end(); I++)
		//for (ILineToCountMap I = m_LineToCountMap.begin(); I != m_LineToCountMap.end(); I++)
			m_LineToCountMap[ I->first ] += I->second;
	}
	string lineStat() {
		string result = "";
		int counter = 0;
		for (ILineToCountMap I = m_LineToCountMap.begin(); I != m_LineToCountMap.end(); I++, counter++) {
			char buf[64];
			sprintf(buf, "%d:%d", I->first, I->second);
			if ( counter > 0 )
				result += ",";
			result += buf;
		}
		return result;
	}
	/*void normalize(double k) {
		m_TotalTime *= k;
		m_RealTime *= k;
	}*/
};

typedef vector<jmethodID> TMethodVector;
typedef TMethodVector::iterator IMethodVector;
class TMethodCallItem {
private:
	bool m_Empty;
	jmethodID m_Method;
	TMethodBillInfo m_Info;
	typedef vector<TMethodCallItem> TMethodCallItemVector;
	typedef TMethodCallItemVector::iterator IMethodCallItemVector;
	TMethodCallItemVector m_Children;
	//
	jlong getRealTime() {
		if ( m_Info.m_RealTime >= 0 )
			return m_Info.m_RealTime;
		jlong childrenTotalTime = 0;
		for (unsigned i = 0; i < m_Children.size(); i++)
			childrenTotalTime += m_Children[i].m_Info.m_TotalTime;
		m_Info.m_RealTime = m_Info.m_TotalTime - childrenTotalTime;
		return m_Info.m_RealTime;
	}
public:
	TMethodCallItem(jmethodID method) {
		init(method);
	}
	TMethodCallItem() {
		m_Empty = true;
	}
	bool isEmpty() {
		return m_Empty;
	}
	void init(jmethodID method) {
		m_Empty = false;
		m_Method = method;
	}
	void clearRealTime() {
		m_Info.m_RealTime = -1;
		for (unsigned i = 0; i < m_Children.size(); i++)
			m_Children[i].clearRealTime();
	}
	void checkRootAndReplace(IMethodVector I, IMethodVector E) {
		if ( I == E )
			return;
		if ( isEmpty() )
			return;
		jmethodID method = *I;
		if ( m_Method == method )
			return;
		//
		m_Children.clear();
		m_Method = method;
		m_Info = TMethodBillInfo();
	}
	void add(IMethodVector I, IMethodVector E, jlong delta) {
		if ( I == E )
			return;
		//
		if ( isEmpty() ) {
			init( *I );
			add(I, E, delta);
			return;
		}
		//
		unsigned i;
		jmethodID method = *I;
		if ( m_Method != method ) {
			printf("m_Method: %s\n", getMethodName(m_Method).c_str());
			printf("method: %s\n", getMethodName(method).c_str());
			printf("Call trace bug.\n");
			exit(0);
		}
		//
		m_Info.m_Count++;
		m_Info.m_TotalTime += delta;
		//
		I++;
		if ( I == E )
			return;
		//
		method = *I;
		int index = -1;
		for (i = 0; i < m_Children.size(); i++) {
			if ( m_Children[i].m_Method == method) {
				index = i;
				break;
			}
		}
		//
		if ( index == -1 ) {
			m_Children.push_back( TMethodCallItem(method) );
			index = m_Children.size() - 1;
		}
		//
		m_Children[index].add(I, E, delta);
	}
	void print(const string &threadName, unsigned level, unsigned maxLevel) {
		if ( isEmpty() )
			return;
		if (level > maxLevel)
			return;
		//
		unsigned i;
		doOut("CallTree|%d|%s|", level, threadName.c_str());
		doOut(
#ifdef WIN32
			"%s|%I64d|%I64d|%I64d\n",
#endif
#ifdef LINUX
			"%s|%qd|%qd|%qd\n",
#endif
			getMethodName(m_Method).c_str(),
			m_Info.m_Count,
			m_Info.m_TotalTime,
			getRealTime());
		doOut("begin children\n");
		for (i = 0; i < m_Children.size(); i++)
			m_Children[i].print(threadName, level + 1, maxLevel);
		doOut("end children\n");
	}
	jlong getTotalTime() {
		return m_Info.m_TotalTime;
	}
	/*void normalize(double k) {
	m_Info.normalize( k );
	for (int i = 0; i < m_Children.size(); i++)
	m_Children[i].normalize(k);
	}*/
};


#ifdef WIN32
	typedef hash_map<jmethodID, TMethodBillInfo, hash2<jmethodID> > TMethodToInfoMap;
#endif
#ifdef LINUX
	typedef map<jmethodID, TMethodBillInfo > TMethodToInfoMap;
#endif
typedef TMethodToInfoMap::iterator IMethodToInfoMap;

class TThreadInfo {
private:
#ifdef WIN32
	HANDLE m_Handle;
#endif
	jlong m_LastTime;
	TMethodToInfoMap m_MethodToInfoMap;
	TMethodCallItem m_Root;
	jlong m_RealTime;
public:
	JNIEnv *m_Env;
	string m_Name;
#ifdef WIN32
	TThreadInfo(JVMPI_Event *event, HANDLE handle) {
		m_Handle = getRealThreadHandle(handle);
		m_Env = event->u.thread_start.thread_env_id;
		m_Name = event->u.thread_start.thread_name;
		m_LastTime = -1;
		m_RealTime = 0;
	}
#endif
#ifdef LINUX
	TThreadInfo(JVMPI_Event *event) {
		m_Env = event->u.thread_start.thread_env_id;
		m_Name = event->u.thread_start.thread_name;
		m_LastTime = -1;
		m_RealTime = 0;
	}
#endif
	void countDirect(const JVMPI_CallTrace &trace) {
		bool isFirst = false;
		int i = 0;
		while ( i < trace.num_frames ) {
			jmethodID method = trace.frames[i].method_id;
			isFirst = true;
			if ( !isVisibleMethod(method) )
				isFirst = false;
			if ( isFirst ) {
				m_MethodToInfoMap[method].m_DirectCount++;
				m_MethodToInfoMap[method].m_LineToCountMap[trace.frames[i].lineno]++;
				break;
			}
			i++;
		}
	}
	void handleTrace(const JVMPI_CallTrace &trace) {
		if ( m_LastTime == -1 ) {
#ifdef WIN32
			m_LastTime = getThreadTimeLong( m_Handle );
#endif
#ifdef LINUX
			m_LastTime = CALL(GetCurrentThreadCpuTime)() / 1000000l;
#endif
		}
#ifdef WIN32
		jlong now = getThreadTimeLong( m_Handle );
#endif
#ifdef LINUX
		jlong now = CALL(GetCurrentThreadCpuTime)() / 1000000l;
#endif
		jlong delta = now - m_LastTime;
		if ( trace.num_frames == 0 ) {
			m_LastTime = now;
			return;
		}
		//
		{
			if ( getMethodName(trace.frames[0].method_id) == "java.lang.Thread.sleep((J)V)" ) {
				m_LastTime = now;
				return;
			}
		}
		//
		if ( m_LastTime != 0 ) {
			m_RealTime = now;
			{
				TMethodVector vtrace;
				for (int i = 0; i < trace.num_frames; i++) {
					jmethodID method = trace.frames[trace.num_frames-i-1].method_id;
					if ( !isVisibleMethod(method) )
						continue;
					vtrace.push_back( method );
				}
				m_Root.checkRootAndReplace(vtrace.begin(), vtrace.end());
				m_Root.add(vtrace.begin(), vtrace.end(), delta);
			}
			//
			countDirect( trace );
			for (int i = 0; i < trace.num_frames; i++) {
				jmethodID method = trace.frames[i].method_id;
				if ( !isVisibleMethod(method) )
					continue;
				m_MethodToInfoMap[method].m_Count++;
				m_MethodToInfoMap[method].m_TotalTime += delta;
   			}
		}
		//
		m_LastTime = now;
	}
	void print() {
		printThreadLine();
		printMethodInfo("Method", m_Name, m_MethodToInfoMap, &m_Root, m_RealTime);
		m_Root.clearRealTime();
		m_Root.print(m_Name, 0, INT_MAX);
	}
	static void printMethodInfo(const char *prefix, string name, TMethodToInfoMap map, TMethodCallItem *root, jlong realTime) {
		int counter = 0;
		for (IMethodToInfoMap I = map.begin(); I != map.end(); I++, counter++) {
			TMethodBillInfo &bill = I->second;
			jlong directTime = 0;
			jlong totalTime = bill.m_TotalTime;
			if ( bill.m_Count > 0 ) {
				directTime = (jlong)(bill.m_TotalTime * bill.m_DirectCount) / bill.m_Count;
			}
			doOut(
#ifdef WIN32
				"%s|%d|%s|%s|%I64d|%I64d|%I64d|%I64d|%s\n",
#endif
#ifdef LINUX
				"%s|%d|%s|%s|%qd|%qd|%qd|%qd|%s\n",
#endif
				prefix,
				counter, 
				name.c_str(), 
				getMethodName(I->first).c_str(), 
				bill.m_Count, 
				bill.m_DirectCount, 
				totalTime, 
				directTime,
				bill.lineStat().c_str());
		}
	}
	void printThreadLine() {
#ifdef WIN32
		doOut("Thread|%s|%I64d\n", m_Name.c_str(), m_RealTime);
#endif
#ifdef LINUX
		doOut("Thread|%s|%qd\n", m_Name.c_str(), m_RealTime);
#endif
	}
	void update(TMethodToInfoMap &map) {
		for (IMethodToInfoMap I = m_MethodToInfoMap.begin(); I != m_MethodToInfoMap.end(); I++) {
			TMethodBillInfo &total = map[ I->first ];
			TMethodBillInfo &bill = I->second;
			total.update( bill );
		}
	}
};

typedef map<JNIEnv*, shared_ptr<TThreadInfo> > TEnvToThreadMap;
typedef TEnvToThreadMap::iterator IEnvToThreadMap;


class TInstanceInfo {
private:
	JVMPI_CallTrace m_Trace;
public:
	jobjectID m_ID;
	jint m_Array;
	jobjectID m_Class;
	jlong m_Size;
	jlong m_BornTime;
	TInstanceInfo() {
		m_Trace.frames = NULL;
		m_BornTime = GetTickCount();
	}
	~TInstanceInfo();
	bool operator==(const TInstanceInfo &inst) const {
		return m_ID == inst.m_ID;
	}
	bool operator==(const jobjectID id) const {
		return m_ID == id;
	}
	JVMPI_CallTrace& getTrace() {
		return m_Trace;
	}
};
TCriticalSection g_CSInstances;
#ifdef WIN32
typedef hash_map<jobjectID, TInstanceInfo, hash2<jobjectID> > TObjectToInstanceInfoMap;
#endif
#ifdef LINUX
typedef map<jobjectID, TInstanceInfo > TObjectToInstanceInfoMap;
#endif
TObjectToInstanceInfoMap g_Instances;
int g_SitesLevel = 10;


class TJVMInfo {
private:
	TEnvToThreadMap m_Threads;
	vector< shared_ptr<TThreadInfo> > m_DeadThreads;
	//
#ifdef WIN32
	hash_map<jobjectID, shared_ptr<TJVMClassInfo>, hash2<jobjectID> > m_IdToClassMap;
#endif
#ifdef LINUX
	map<jobjectID, shared_ptr<TJVMClassInfo> > m_IdToClassMap;
#endif
	TIdToMethodMap m_IdToMethodMap;
	TMethodIdToClassMap m_MethodIdToClassMap;
	//
	vector<JVMPI_CallFrame> m_CallFrames;
	deque<int> m_FreeCallFrames;
	vector<char> m_CallFrameStatuses;
	unsigned m_LastCallFrame;
	//
	static TJVMInfo *m_Instance;
	TJVMInfo() {
		//m_CallFrames.reserve(10000 * 64);
		//m_CallFrameStatuses.reserve(10000 * 64);
		m_LastCallFrame = 0;
	}
public:
	static TJVMInfo* instance() {
		if ( m_Instance == 0 )
			m_Instance = new TJVMInfo();
		return m_Instance;
	}
	void addClass(JVMPI_Event *event) {
		jobjectID objectID = event->u.class_load.class_id;
		if ( m_IdToClassMap.count(objectID) == 1 )
			return;
		shared_ptr<TJVMClassInfo> info( new TJVMClassInfo(event) );
		m_IdToClassMap[objectID] = info;
		info->addMethods(m_IdToMethodMap, m_MethodIdToClassMap, info);
		//printf("add class: %s\n", info->m_Name.c_str());
	}
	void findClass(jobjectID objectID) {
		if (m_IdToClassMap.count(objectID) > 0)
			return;
		//printf("send request: %X\n", objectID);
		//fflush(stdout);
		//CALL(DisableGC)();
		try {
			CALL(RequestEvent)(JVMPI_EVENT_CLASS_LOAD, objectID);
		}
		catch ( ... ) {
			//printf("crash: %X\n", objectID);
			//fflush(stdout);
		}
		//CALL(EnableGC)();
		//printf("done\n");
		//fflush(stdout);
	}
    shared_ptr<TJVMClassInfo> getClass(jobjectID objectID) {
		if ((objectID == 0) || (m_IdToClassMap.count(objectID) == 0))
			return shared_ptr<TJVMClassInfo>(new TJVMClassInfo());
		return m_IdToClassMap[objectID];
    }
    void moveClass(jobjectID objectID, jobjectID newObjectID) {
		if ((objectID == 0) || (m_IdToClassMap.count(objectID) == 0))
			return;
		shared_ptr<TJVMClassInfo> info = m_IdToClassMap[objectID];
		info->m_ID = newObjectID;
		//printf("move class '%s' from %p to %p\n", info->m_Name.c_str(), objectID, newObjectID);
		//fflush(stdout);
		m_IdToClassMap.erase(objectID);
		m_IdToClassMap[newObjectID] = info;
		// change class id in instances
		for (TObjectToInstanceInfoMap::iterator I = g_Instances.begin(); I != g_Instances.end(); I++) {
			TInstanceInfo &info = I->second;
			if ( info.m_Class == objectID )
				info.m_Class = newObjectID;
		}
    }
	bool hasMethod(jmethodID methodID) {
		return m_IdToMethodMap.count(methodID) != 0;
	}
	bool hasClass(jmethodID methodID) {
		return m_MethodIdToClassMap.count(methodID) != 0;
	}
	shared_ptr<TJVMMethodInfo> getMethod(jmethodID methodID) {
		return m_IdToMethodMap[methodID];
	}
	string getMethodName(jmethodID methodID) {
		if ( !hasMethod(methodID) || !hasClass(methodID) ) {
			/*printf("unknown class or method: %p\n", methodID);
			fflush(stdout);*/
			return "???";
		}
		shared_ptr<TJVMClassInfo> clazz = getClass( m_MethodIdToClassMap[methodID]->m_ID );
		shared_ptr<TJVMMethodInfo> method = getMethod(methodID);
		return clazz->m_Name + "." + method->m_Name + "(" + method->m_Signature + ")";
		/*shared_ptr<TJVMMethodInfo> method = getMethod(methodID);
		return method->m_Name;*/
	}
	void addThread(JVMPI_Event *event) {
		JNIEnv *env = event->u.thread_start.thread_env_id;
		if ( m_Threads.count(env) == 0 ) {
#ifdef WIN32
			m_Threads[env] = shared_ptr<TThreadInfo>( new TThreadInfo(event, GetCurrentThread()) );
#endif
#ifdef LINUX
			m_Threads[env] = shared_ptr<TThreadInfo>( new TThreadInfo(event) );
#endif
		}
	}
	shared_ptr<TThreadInfo> getThread(JNIEnv *env) {
		return m_Threads[env];
	}
	bool hasThread(JNIEnv *env) {
		return m_Threads.count(env) != 0;
	}
	vector<JNIEnv*> getThreadEnvs() {
		vector<JNIEnv*> result;
		map<JNIEnv*, shared_ptr<TThreadInfo> >::iterator I = m_Threads.begin();
		while ( I != m_Threads.end() ) {
			result.push_back( I->first );
			I++;
		}
		return result;
	}
	void remThread(JVMPI_Event *event) {
		m_DeadThreads.push_back( m_Threads[event->env_id] );
		m_Threads.erase(event->env_id);
	}
	// print
	void print() {
		// print total methods info
		{
			TMethodToInfoMap map;
			{
				for (IEnvToThreadMap I = m_Threads.begin(); I != m_Threads.end(); I++)
					I->second->update( map );
			}
			{
				for (unsigned i = 0; i < m_DeadThreads.size(); i++)
					m_DeadThreads[i]->update( map );
			}
			TThreadInfo::printMethodInfo("AllMethods", "all", map, NULL, 0);
		}
		// print each thread info
		{
			for (IEnvToThreadMap I = m_Threads.begin(); I != m_Threads.end(); I++)
				I->second->print();
			for (unsigned i = 0; i < m_DeadThreads.size(); i++)
				m_DeadThreads[i]->print();
		}
	}
	void printThreadInfo() {
		for (IEnvToThreadMap I = m_Threads.begin(); I != m_Threads.end(); I++)
			I->second->printThreadLine();
		for (unsigned i = 0; i < m_DeadThreads.size(); i++)
			m_DeadThreads[i]->printThreadLine();
	}
	JVMPI_CallFrame* getCallFrame() {
		/*if ( m_CallFrames.empty() )
			m_CallFrames.reserve( g_SitesLevel );
		JVMPI_CallFrame* result = &m_CallFrames[0];
		return result;*/
		int index = -1;
		if ( !m_FreeCallFrames.empty() ) {
			index = m_FreeCallFrames.front();
			m_FreeCallFrames.pop_front();
		} else {
			if ( m_LastCallFrame >= m_CallFrameStatuses.capacity() ) {
				/*printf("do increment\n");
				fflush(stdout);*/
				//
				{
					for (TObjectToInstanceInfoMap::iterator I = g_Instances.begin(); I != g_Instances.end(); I++) {
						TInstanceInfo &info = I->second;
						if ( info.getTrace().frames == NULL )
							continue;
						if ( (info.getTrace().num_frames != 0) && (info.getTrace().frames != 0) && (info.getTrace().frames[0].method_id == 0) ) {
							printf("dump: %d\n", g_Instances.size());
							fflush(stdout);
							exit(1);
						}
					}
				}
				//
				JVMPI_CallFrame *old_p = &m_CallFrames[0];
				//JVMPI_CallFrame s;
				//m_CallFrames.push_back(s);
				//m_CallFrameStatuses.push_back(0);
				int increment = 10000;
				/*printf("size: %d %d\n", m_CallFrames.size(), g_Instances.size());
				fflush(stdout);
				if ( m_CallFrames.size() > 0 ) {
					printf("before: %p\n", m_CallFrames[0].method_id);
					fflush(stdout);
				}*/
				m_CallFrames.reserve(m_CallFrames.capacity() + g_SitesLevel * increment);
				m_CallFrames.resize( m_CallFrames.capacity() );
				/*printf("reserve: %p\n", m_CallFrames[0].method_id);
				fflush(stdout);*/
				m_CallFrameStatuses.reserve(m_CallFrameStatuses.capacity() + increment);
				m_CallFrameStatuses.resize( m_CallFrameStatuses.capacity() );
				/*for (unsigned i = m_LastCallFrame; i < m_CallFrameStatuses.capacity(); i++)
					m_CallFrameStatuses[i] = 0;*/
				//
				JVMPI_CallFrame *new_p = &m_CallFrames[0];
				int i = 0;
				for (TObjectToInstanceInfoMap::iterator I = g_Instances.begin(); I != g_Instances.end(); I++,i++) {
					TInstanceInfo &info = I->second;
					if ( info.getTrace().frames == NULL )
						continue;
					info.getTrace().frames = new_p + (info.getTrace().frames - old_p);
					//printf("fames: %p\n", info.m_Trace.frames);
					//fflush(stdout);
					/*if ( (info.getTrace().num_frames != 0) && (info.getTrace().frames != 0) && (info.getTrace().frames[0].method_id == 0) ) {
						printf("move: %d/%d when %p to %p\n", i, g_Instances.size(), old_p, new_p);
						fflush(stdout);
						exit(1);
					}*/
					/*jmethodID oldID = 0;
					printf("frames: %p\n", info.m_Trace.frames);
					fflush(stdout);
					if (info.m_Trace.frames != 0)
						oldID = info.m_Trace.frames[0].method_id;
					printf("move: %08X\n", oldID);
					fflush(stdout);*/
					/*info.m_Trace.frames = new_p + (info.m_Trace.frames - old_p);
					if ( (oldID != 0) && (old_p != 0) && (info.m_Trace.frames != 0) && (info.m_Trace.frames[0].method_id == 0) ) {
						printf("move: %d when %p to %p\n", g_Instances.size(), old_p, new_p);
						//printf("move: %08X\n", oldID);
						fflush(stdout);
						exit(1);
					}*/
				}
			}
			index = m_LastCallFrame;
			m_LastCallFrame++;
		}
		JVMPI_CallFrame* result = &m_CallFrames[index * g_SitesLevel];
		//m_CallFrameStatuses[index] = 1;
		return result;
	}
	void freeCallFrame(JVMPI_CallFrame *p) {
		unsigned index = (p - &m_CallFrames[0]) / g_SitesLevel;
		/*printf("free: %p %p %d %d\n", p,  &m_CallFrames[0], p - &m_CallFrames[0], index);
		printf("????: %p %p\n", &m_CallFrames[0], &m_CallFrames[1]);
		fflush(stdout);*/
		m_CallFrameStatuses[index] = 0;
		m_FreeCallFrames.push_back(index);
	}
	void packCallFrames() {
		/*for (TObjectToInstanceInfoMap::iterator I = g_Instances.begin(); I != g_Instances.end(); I++) {
			TInstanceInfo &info = I->second;
			if ( info.m_Class == objectID )
				info.m_Class = newObjectID;
		}*/
	}
};
TJVMInfo* TJVMInfo::m_Instance;

TInstanceInfo::~TInstanceInfo() {
	/*if ( m_Trace.frames != NULL )
		delete [] m_Trace.frames;*/
	if ( m_Trace.frames != NULL )
		TJVMInfo::instance()->freeCallFrame(m_Trace.frames);
}


string
getMethodName(jmethodID methodID) {
	 return TJVMInfo::instance()->getMethodName(methodID);
}


class TTypeInfo {
public:
	jint m_Array;
	jobjectID m_Class;
	jlong m_Count;
	jlong m_Size;
	TTypeInfo() {
		m_Count = 0;
		m_Size = 0;
    }
};
#ifdef WIN32
	typedef hash_map<jlong, TTypeInfo, hash2<jlong> > TClassToTypeMap;
	typedef hash_map<jint, TClassToTypeMap> TArrayToClassesMap;
#endif
#ifdef LINUX
	typedef map<jlong, TTypeInfo > TClassToTypeMap;
	typedef map<jint, TClassToTypeMap> TArrayToClassesMap;
#endif
typedef map<jmethodID,jlong> TMethodToSizeMap;
TArrayToClassesMap g_Types;

class CompareClassBySize : public greater<jlong> {
private:
	TClassToTypeMap &m_ClassToType;
public:
	CompareClassBySize(TClassToTypeMap &classToType) : m_ClassToType(classToType) {
	}
	bool operator()(jlong id1, jlong id2) const {
		return m_ClassToType[id1].m_Size > m_ClassToType[id2].m_Size;
	}
};

class CompareSitesBySize : public greater<jlong> {
private:
	TMethodToSizeMap &m_MethodToSize;
public:
	CompareSitesBySize(TMethodToSizeMap &map) : m_MethodToSize(map) {
	}
	bool operator()(jmethodID id1, jmethodID id2) const {
		return m_MethodToSize[id1] > m_MethodToSize[id2];
	}
};

string
getTypeName(shared_ptr<TJVMClassInfo> info, jint array) {
	if ( array == JVMPI_CLASS )
		return info->m_Name + "[]";
	if ( array == JVMPI_BOOLEAN )
		return "boolean[]";
	if ( array == JVMPI_BYTE )
		return "byte[]";
	if ( array == JVMPI_CHAR )
		return "char[]";
	if ( array == JVMPI_SHORT )
		return "short[]";
	if ( array == JVMPI_INT )
		return "int[]";
	if ( array == JVMPI_LONG )
		return "long[]";
	if ( array == JVMPI_FLOAT )
		return "float[]";
	if ( array == JVMPI_DOUBLE )
		return "double[]";
	return info->m_Name;
	//return "???";
}

JavaVM *jvm;
int hookMe = 0;

jclass g_TraceClass = 0;
jmethodID g_InstrumentClassMethod = 0;
jmethodID g_ForceWriteMethod = 0;
//

int
initCallbacks(JNIEnv *env) {
	if ( g_TraceClass != 0 )
		return 1;
	//
	g_TraceClass = env->FindClass("ru/nlmk/eclipse/plugins/profiler/trace/Trace");
	if ( g_TraceClass == 0 )
		return 0;
	g_InstrumentClassMethod = env->GetStaticMethodID(g_TraceClass, "instrumentClass", "([B)[B");
	if ( g_InstrumentClassMethod == 0 )
		return 0;
	g_ForceWriteMethod = env->GetStaticMethodID(g_TraceClass, "forceWriteStatistictToFrontEnd", "()V");
	if ( g_ForceWriteMethod == 0 )
		return 0;
	return 1;
}

bool initDone = false;
int cnt = 0;
static void sample_thread_loop(void *p);
TCriticalSection g_MonitorCS;
jobjectID g_MonitorObj;


int lastSiteLevel = -1;	

void notifyEvent(JVMPI_Event *event) {
	JNIEnv *env;
	jvm->GetEnv((void **)&env, JNI_VERSION_1_2);
	//
	switch(event->event_type) {
	case JVMPI_EVENT_OBJECT_DUMP | JVMPI_REQUESTED_EVENT:
		{
			printf("JVMPI_EVENT_OBJECT_DUMP\n");
			char *begin = event->u.object_dump.data;
			int type = *(begin++);
			jobjectID id = (jobjectID)*((int*)(begin));
			begin+=4;
			jobjectID classID = (jobjectID)*((int*)(begin));
			begin+=4;
			printf("type: %02X %p %p %s\n", type, id, classID, TJVMInfo::instance()->getClass(classID)->m_Name.c_str());
			fflush(stdout);
		}
		break;
	case JVMPI_EVENT_JVM_INIT_DONE:
		initDone = true;
		return;
	case JVMPI_EVENT_JVM_SHUT_DOWN:
        initCallbacks( env );
		{
			if ( !initDone || (g_TraceClass == 0) || (g_ForceWriteMethod == 0) )
				return;
			env->CallStaticVoidMethod(g_TraceClass, g_ForceWriteMethod);
		}
		initDone = false;
		break;
	case JVMPI_EVENT_HEAP_DUMP | JVMPI_REQUESTED_EVENT:
		{
			printf("DUMP: %d\n", event->u.heap_dump.end - event->u.heap_dump.begin);
			fflush(stdout);
		}
		break;
	case JVMPI_EVENT_THREAD_START:
	case JVMPI_EVENT_THREAD_START | JVMPI_REQUESTED_EVENT:
		{
			TJVMInfo::instance()->addThread(event);
		}
		break;
	case JVMPI_EVENT_THREAD_END:
		{
			TJVMInfo::instance()->remThread(event);
		}
		break;
	case JVMPI_EVENT_CLASS_LOAD: 
	case JVMPI_EVENT_CLASS_LOAD | JVMPI_REQUESTED_EVENT: 
		/*if ( !initDone )
			return;*/
		{
			TCSLocker CSL(g_CSInstances);
			TJVMInfo::instance()->addClass(event);
		}
		break;
	case JVMPI_EVENT_CLASS_LOAD_HOOK:
		event->u.class_load_hook.new_class_data = event->u.class_load_hook.class_data;
		event->u.class_load_hook.new_class_data_len = event->u.class_load_hook.class_data_len;
		if ( !initDone )
			return;
		{
			event->u.class_load_hook.new_class_data = event->u.class_load_hook.class_data;
			event->u.class_load_hook.new_class_data_len = event->u.class_load_hook.class_data_len;
			if ( hookMe == 1 )
				return;
			//
			hookMe = 1;
			if ( initCallbacks( env ) == 0 ) {
				hookMe = 0;
				return;
			}
			if ( (g_TraceClass == 0) || (g_InstrumentClassMethod == 0) )
				return;
			jbyteArray buf = env->NewByteArray(event->u.class_load_hook.class_data_len);
			env->SetByteArrayRegion(buf, 0, event->u.class_load_hook.class_data_len, (jbyte *)event->u.class_load_hook.class_data);
			jbyteArray buf_new;
#ifdef WIN32
			try {
				buf_new = (jbyteArray)env->CallStaticObjectMethod(g_TraceClass, g_InstrumentClassMethod, buf);
			}
			catch (...) {
				hookMe = 0;
				return;
			}
#endif
#ifdef LINUX
			buf_new = (jbyteArray)env->CallStaticObjectMethod(g_TraceClass, g_InstrumentClassMethod, buf);
#endif
			hookMe = 0;
			//
			int new_len = env->GetArrayLength(buf_new);
			event->u.class_load_hook.new_class_data_len = new_len;
			event->u.class_load_hook.new_class_data = (unsigned char*)event->u.class_load_hook.malloc_f(new_len);
			env->GetByteArrayRegion(buf_new, 0, new_len, (jbyte*)event->u.class_load_hook.new_class_data);
		}
		break;
	case JVMPI_EVENT_OBJECT_ALLOC:
		if ( !initDone )
			return;
		{
			jobjectID clazz = event->u.obj_alloc.class_id;
			jobjectID obj = event->u.obj_alloc.obj_id;
			jint size = event->u.obj_alloc.size;
			jint is_array = event->u.obj_alloc.is_array;
			//
			{
				TCSLocker CSL(g_CSInstances);
				//
#ifdef WIN32
				/*if ( clazz != 0 )
					TJVMInfo::instance()->findClass( clazz );*/
#endif
				//
				TInstanceInfo &instInfo = g_Instances[obj];
				instInfo.m_Array = is_array;
				instInfo.m_Class = clazz;
				instInfo.m_Size = size;
				//
				if ( g_SitesLevel > 0 ) {
					//JVMPI_CallTrace trace;
					//instInfo.getTrace().frames = new JVMPI_CallFrame[g_SitesLevel];
					instInfo.getTrace().frames = TJVMInfo::instance()->getCallFrame();
					//printf("instInfo.m_Trace.frames: %p\n", instInfo.m_Trace.frames);
					//fflush(stdout);
					instInfo.getTrace().num_frames = 0;
					instInfo.getTrace().env_id = event->env_id;
					//CALL(GetCallTrace)(&instInfo.getTrace(), g_SitesLevel);
					//printf("level: %d  num_frames: %d\n", g_SitesLevel, instInfo.getTrace().num_frames);
					int level = lastSiteLevel;
					if (level == -1)
						level = 8;
					while ( true ) {
						CALL(GetCallTrace)(&instInfo.getTrace(), g_SitesLevel);
						if (instInfo.getTrace().num_frames == level)
							level += 4;
						else
							break;
					}
					if ( instInfo.getTrace().num_frames < level / 2 )
						level /= 2;
					lastSiteLevel = level;
					//instInfo.m_Trace = trace;
				}
			}
			//checkAndPrintInstances();
			//
		}
		break;
	case JVMPI_EVENT_OBJECT_MOVE:
		if ( !initDone )
			return;
		{
			jobjectID obj = event->u.obj_move.obj_id;
			jobjectID new_obj = event->u.obj_move.new_obj_id;
			/*if ( obj == g_MonitorObj ) {
				g_MonitorObj = new_obj;
				printf("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!: %p %p\n", obj, new_obj);
				fflush(stdout);
			}*/
			//TCSLocker CSL(g_CSInstances);
			TJVMInfo::instance()->moveClass(obj, new_obj);
			/*TInstanceInfo instInfo = g_Instances[obj];
			g_Instances.erase(obj);
			g_Instances[new_obj] = instInfo;*/
		}
		break;
	case JVMPI_EVENT_OBJECT_FREE:
		if ( !initDone )
			return;
		{
			jobjectID obj = event->u.obj_free.obj_id;
			TCSLocker CSL(g_CSInstances);
			g_Instances.erase( obj );
		}
		break;
	case JVMPI_EVENT_GC_FINISH:
		{
			/*printf("JVMPI_EVENT_GC_FINISH\n");
			fflush(stdout);*/
		}
		break;
	case JVMPI_EVENT_MONITOR_WAIT:
		if ( !initDone )
			return;
		{
			TCSLocker csl(g_MonitorCS);
			//printf("JVMPI_EVENT_MONITOR_WAIT: %d\n", event->u.monitor_wait.timeout);
			//printf("JVMPI_EVENT_MONITOR_WAIT: %p %d\n", event->u.monitor_wait.object, event->u.monitor_wait.timeout);
			/*if ( !TJVMInfo::instance()->hasThread(event->env_id) )
				break;*/
			string name = TJVMInfo::instance()->getThread(event->env_id)->m_Name;
			/*if ( name != "main" )
				break;*/
			printf("wait thread: %s\n", name.c_str());
			g_MonitorObj = event->u.monitor_wait.object;
			CALL(RequestEvent)(JVMPI_EVENT_OBJECT_DUMP, g_MonitorObj);
			//printf("event->env_id: %p\n", event->env_id);
			//printf("JVMPI_EVENT_MONITOR_WAIT: %p %p %d %p\n", event->env_id, g_MonitorObj, event->u.monitor_wait.timeout, jvmpi_interface->jobjectID2jobject(event->u.monitor_wait.object));
			/*printf("JVMPI_EVENT_MONITOR_WAIT: %p %p %d\n", event->env_id, g_MonitorObj, event->u.monitor_wait.timeout);
			//printf("JVMPI_EVENT_MONITOR_WAIT: %p %p %d\n", event->env_id, CALL(jobjectID2jobject)(event->u.monitor_wait.object), event->u.monitor_wait.timeout);
			//
			vector<JNIEnv*> envs = TJVMInfo::instance()->getThreadEnvs();
			for (unsigned i = 0; i < envs.size(); i++) {
				JNIEnv *env = envs[i];
				printf("%s: %d\n", TJVMInfo::instance()->getThread(env)->m_Name.c_str(), CALL(GetThreadStatus)(env));
			}*/
			//
			fflush(stdout);
		}
		break;
	case JVMPI_EVENT_MONITOR_WAITED:
		if ( !initDone )
			return;
		{
			TCSLocker csl(g_MonitorCS);
			//printf("JVMPI_EVENT_MONITOR_WAITED: %d\n", event->u.monitor_wait.timeout);
			//printf("JVMPI_EVENT_MONITOR_WAITED: %p %d\n", event->u.monitor_wait.object, event->u.monitor_wait.timeout);
			/*if ( !TJVMInfo::instance()->hasThread(event->env_id) )
				break;*/
			string name = TJVMInfo::instance()->getThread(event->env_id)->m_Name;
			/*if ( name != "main" )
				break;*/
			printf("waited thread: %s\n", name.c_str());
			g_MonitorObj = event->u.monitor_wait.object;
			CALL(RequestEvent)(JVMPI_EVENT_OBJECT_DUMP, g_MonitorObj);
			/*g_MonitorObj = event->u.monitor_wait.object;
			printf("JVMPI_EVENT_MONITOR_WAITED: %p %p %d\n", event->env_id, g_MonitorObj, event->u.monitor_wait.timeout);
			//printf("JVMPI_EVENT_MONITOR_WAITED: %p %p %d\n", event->env_id, CALL(jobjectID2jobject)(event->u.monitor_wait.object), event->u.monitor_wait.timeout);
			//
			vector<JNIEnv*> envs = TJVMInfo::instance()->getThreadEnvs();
			for (unsigned i = 0; i < envs.size(); i++) {
				JNIEnv *env = envs[i];
				printf("%s: %d\n", TJVMInfo::instance()->getThread(env)->m_Name.c_str(), CALL(GetThreadStatus)(env));
			}*/
			//
			fflush(stdout);
		}
		break;
	}
}


int g_PrintLock = 0;
JVMPI_RawMonitor g_SamplingMonitor;


/*static void
heap_thread_loop(void *p) {
	while ( true ) {
		printf("dump...\n");
		fflush(stdout);
		//
		JVMPI_HeapDumpArg arg;
		arg.heap_dump_level = JVMPI_DUMP_LEVEL_2;
		CALL(RequestEvent)(JVMPI_EVENT_HEAP_DUMP, &arg);
		//
		Sleep( 1000 );
	}
}*/

int lastPrint = 0;
static void
sample_thread_loop(void *p) {
    JNIEnv *my_env;
    jvm->GetEnv((void **)&my_env, JNI_VERSION_1_2);
	JVMPI_RawMonitor g_Monitor = CALL(RawMonitorCreate)("___Notify monitor");
	g_SamplingMonitor = CALL(RawMonitorCreate)("___Sampling monitor");
	//Sleep(1000);
	//
	/*jobject runtime;
	jmethodID freeMemoryMethod;
	jmethodID totalMemoryMethod;
	{
		jclass clazz = my_env->FindClass("Ljava/lang/Runtime;");
		jmethodID getRuntimeMethod = my_env->GetStaticMethodID(clazz, "getRuntime", "()Ljava/lang/Runtime;");
		runtime = my_env->CallStaticObjectMethod(clazz, getRuntimeMethod);
		freeMemoryMethod = my_env->GetMethodID(clazz, "freeMemory", "()J");
		totalMemoryMethod = my_env->GetMethodID(clazz, "totalMemory", "()J");
	}*/
	//
	//SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
	//SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);
	//
	/*{
		UINT wTimerRes;
		TIMECAPS tc;
		timeGetDevCaps(&tc, sizeof(TIMECAPS));
		wTimerRes = min(max(tc.wPeriodMin, 1), tc.wPeriodMax);
		timeBeginPeriod(wTimerRes);	
	}*/
	//
	//jlong lastTime = getTime();
    CALL(RawMonitorEnter)(g_Monitor);
	while ( true ) {
		//Sleep(0);
		CALL(RawMonitorWait)(g_Monitor, (jlong)1);
		/*while ( g_PrintLock )
		Sleep(1);
		g_PrintLock = 1;*/
		//
		/*jlong nowTime = getTime();
		printf("%I64d\n", (nowTime - lastTime));
		fflush(stdout);
		lastTime = nowTime;*/
		//
		//eventCount++;
		vector<JNIEnv*> threads;
		unsigned i;
		//
		vector<JNIEnv*> envs = TJVMInfo::instance()->getThreadEnvs();
		for (i = 0; i < envs.size(); i++) {
			JNIEnv *env = envs[i];
			if ( (env != my_env) && ((CALL(GetThreadStatus)(env) & (~JVMPI_THREAD_INTERRUPTED)) == JVMPI_THREAD_RUNNABLE) ) {
				/*if (TJVMInfo::instance()->getThread(env)->m_Name != "main")
					continue;*/
				threads.push_back( env );
				/*printf("%s\n", TJVMInfo::instance()->getThread(env)->m_Name);
				fflush(stdout);*/
				//CALL(SuspendThread)(env);
				//CALL(ResumeThread)(env);
			}
		}
		//
		CALL(DisableGC)();
		{
			int max_level = 1000;
			JVMPI_CallTrace trace;
			trace.frames = new JVMPI_CallFrame[max_level];
			for (i = 0; i < threads.size(); i++) {
				TCSLocker csl( g_SamplingCS );
				JNIEnv *env = threads[i];
				trace.num_frames = 0;
				trace.env_id = env;
				//CALL(RawMonitorEnter)(g_SamplingMonitor);
				CALL(SuspendThread)(env);
				CALL(GetCallTrace)(&trace, max_level);
				CALL(ResumeThread)(env);
				//CALL(RawMonitorExit)(g_SamplingMonitor);
				//printf("%d\n", trace.num_frames);
				//fflush(stdout);
				shared_ptr<TThreadInfo> thread = TJVMInfo::instance()->getThread(env);
				if ( thread.get() != NULL )
					thread->handleTrace(trace);
			}
			delete [] trace.frames;
		}
		CALL(EnableGC)();
		//
		/*for (i = 0; i < threads.size(); i++) {
			JNIEnv *env = threads[i];
			CALL(ResumeThread)(env);
		}*/
		//g_PrintLock = 0;
	}
}


class TAllocationSite {
private:
	jmethodID m_Method;
	jint m_Line;
	vector<TAllocationSite> m_Children;
	int m_Count;
	jlong m_Size;
	int m_DirectCount;
	jlong m_DirectSize;
public:
	TAllocationSite() {
		m_Method = 0;
		m_Line = 0;
		m_Count = 0;
		m_Size = 0;
		m_DirectCount = 0;
		m_DirectSize = 0;
	}
	TAllocationSite(JVMPI_CallFrame &frame) {
		m_Method = frame.method_id;
		m_Line = frame.lineno;
		m_Count = 0;
		m_Size = 0;
		m_DirectCount = 0;
		m_DirectSize = 0;
	}
	bool operator==(JVMPI_CallFrame &frame) const {
		//return (m_Method == frame.method_id) && (m_Line == frame.lineno);
		return (m_Method == frame.method_id);
	}
	void add(JVMPI_CallFrame *frames, int pos, jlong size) {
		m_Count++;
		m_Size += size;
		if (pos == 0) {
			m_DirectCount++;
			m_DirectSize += size;
			return;
		}
		pos--;
		for (unsigned i = 0; i < m_Children.size(); i++)
			if (m_Children[i] == frames[pos]) {
				m_Children[i].add(frames, pos, size);
				return;
			}
		TAllocationSite child(frames[pos]);
		child.add(frames, pos, size);
		m_Children.push_back( child );
	}
	void print(string &s, int level) {
		//char buf[1024 * 16];
		char *buf = new char[1024 * 16];
#ifdef WIN32
		sprintf(buf, "%s %d %I64d %d %I64d\n", TJVMInfo::instance()->getMethodName(m_Method).c_str(), m_Count, m_Size, m_DirectCount, m_DirectSize);
#endif
#ifdef LINUX
		sprintf(buf, "%s %d %qd %d %qd\n", TJVMInfo::instance()->getMethodName(m_Method).c_str(), m_Count, m_Size, m_DirectCount, m_DirectSize);
#endif
		s += buf;
		delete [] buf;
		//
		s += "+\n";
		for (unsigned i = 0; i < m_Children.size(); i++)
			m_Children[i].print(s, level + 1);
		s += "-\n";
	}
};

extern "C" { 
	JNIEXPORT jint JNICALL JVM_OnLoad(JavaVM *vm, char *options, void *reserved) {
		jvm = vm;
		jvm->GetEnv((void **)&jvmpi_interface, JVMPI_VERSION_1);
		//
		if (options == NULL) {
			printf("Argument required after -XrunProfilerDLL, for example  -XrunProfilerDLL:1\n");
			fflush(stdout);
			exit(1);
		}
		//
		vector<string> args;
		split(options, ",", back_inserter(args));
		//
		if (args.size() >= 2)
			g_SitesLevel = atoi(args[1].c_str());
		//
		bool useSampling = false;
		if (args.size() >= 3)
			useSampling = atoi(args[2].c_str()) == 1;
		//
		int param = atoi(args[0].c_str());
		/*printf("param: %d\n", param);
		fflush(stdout);*/
		g_Instances.clear();
		// 
		jvmpi_interface->NotifyEvent = notifyEvent;
		jvmpi_interface->EnableEvent(JVMPI_EVENT_JVM_INIT_DONE, NULL);
		jvmpi_interface->EnableEvent(JVMPI_EVENT_CLASS_LOAD, NULL);
		jvmpi_interface->EnableEvent(JVMPI_EVENT_THREAD_START, NULL);
		jvmpi_interface->EnableEvent(JVMPI_EVENT_THREAD_END, NULL); 
		//jvmpi_interface->EnableEvent(JVMPI_EVENT_OBJECT_MOVE, NULL); 
		if ((param & 0x01) == 0x01) {
			jvmpi_interface->EnableEvent(JVMPI_EVENT_CLASS_LOAD_HOOK, NULL);
		}
		if ((param & 0x02) == 0x02) {
			//jvmpi_interface->EnableEvent(JVMPI_EVENT_GC_FINISH, NULL); 
			jvmpi_interface->EnableEvent(JVMPI_EVENT_OBJECT_ALLOC, NULL); 
			jvmpi_interface->EnableEvent(JVMPI_EVENT_OBJECT_MOVE, NULL); 
			jvmpi_interface->EnableEvent(JVMPI_EVENT_OBJECT_FREE, NULL); 
		}
		if ((param & 0x04) == 0x04) {
			jvmpi_interface->EnableEvent(JVMPI_EVENT_MONITOR_WAIT, NULL); 
			jvmpi_interface->EnableEvent(JVMPI_EVENT_MONITOR_WAITED, NULL); 
			jvmpi_interface->EnableEvent(JVMPI_EVENT_OBJECT_MOVE, NULL); 
		}
		jvmpi_interface->EnableEvent(JVMPI_EVENT_JVM_SHUT_DOWN, NULL);
		//
		//CALL(CreateSystemThread)("NLMK HEAP dumper", JVMPI_MAXIMUM_PRIORITY, heap_thread_loop);
		if (useSampling)
			CALL(CreateSystemThread)("NLMK CPU sampler", JVMPI_MAXIMUM_PRIORITY, sample_thread_loop);
		//
		return JNI_OK;
	}
	
	JNIEXPORT jlong JNICALL Java_ru_nlmk_eclipse_plugins_profiler_trace_Trace_getCurrentThreadTime(JNIEnv *env, jclass that, jint method) {
		if (method == 1) {
			if ( jvmpi_interface == NULL )
				return 0;
#ifdef WIN32
			return CALL(GetCurrentThreadCpuTime)() / 1000000l;
#endif
#ifdef LINUX
			return CALL(GetCurrentThreadCpuTime)() / 1000l;
#endif
		}
		return 0;
	}
	JNIEXPORT void JNICALL Java_ru_nlmk_eclipse_plugins_profiler_trace_Trace_nativeEnter(JNIEnv *env, jclass that, int threadId, int method) {
	}
	JNIEXPORT void JNICALL Java_ru_nlmk_eclipse_plugins_profiler_trace_Trace_nativeLeave(JNIEnv *env, jclass that, int threadId, int method) {
	}
	JNIEXPORT void JNICALL Java_ru_nlmk_eclipse_plugins_profiler_trace_Trace_setNativeFilters(JNIEnv *env, jclass that, jbyteArray jstr) {
		/*printf("setNativeFilters\n");
		fflush(stdout);*/
		int len = env->GetArrayLength(jstr);
		char *buf = new char[len + 1];
		env->GetByteArrayRegion(jstr, 0, len, (jbyte*)buf);
		split(string(buf), ",", back_inserter(g_Filters));
		/*for (unsigned i = 0; i < g_Filters.size(); i++)
			printf("%s\n", g_Filters[i].c_str());
		fflush(stdout);*/
	}
	//JNIEXPORT jint JNICALL Java_ru_nlmk_eclipse_plugins_profiler_trace_Trace_getHeapInfo(JNIEnv *env, jclass that, jbyteArray res, jlong liveTime) {
	JNIEXPORT jbyteArray JNICALL Java_ru_nlmk_eclipse_plugins_profiler_trace_Trace_getHeapInfo(JNIEnv *env, jclass that, jlong liveTime) {
		string s = "";
		{
			TCSLocker CSL(g_CSInstances);
			//
			//printf("getHeapInfo: 0\n"); fflush(stdout);
			set<jlong> uniqueTypes;
			vector<jlong> types;
			TClassToTypeMap class2type;
			for (TObjectToInstanceInfoMap::iterator I = g_Instances.begin(); I !=g_Instances.end(); I++) {
				TInstanceInfo &info = I->second;
				if ( GetTickCount() - info.m_BornTime < liveTime )
					continue;
				jobjectID clazz = info.m_Class;
				jint array = info.m_Array;
				jlong id = ((jlong)array) << 32 | (unsigned int)clazz;
				if ( uniqueTypes.count(id) == 0 ) {
					uniqueTypes.insert( id );
					types.push_back( id );
				}
				TTypeInfo &type = class2type[ id ];
				type.m_Array = array;
				type.m_Class = clazz;
				type.m_Count++;
				type.m_Size += info.m_Size;
			}
			//printf("getHeapInfo: 1\n"); fflush(stdout);
			sort(types.begin(), types.end(), CompareClassBySize(class2type));
			//printf("getHeapInfo: 2\n"); fflush(stdout);
			//
			for (unsigned i = 0; i < types.size(); i++) {
				jint array = (jint)(types[i] >> 32) & 0xFF;
				jobjectID clazz = (jobjectID)(int)(types[i] & 0xFFFFFFFF);
				TTypeInfo &type = class2type[ types[i] ];
				//
				if (type.m_Size < 1024)
					break;
				//
				shared_ptr<TJVMClassInfo> info = TJVMInfo::instance()->getClass(clazz);
				char buf[1024];
				string name = getTypeName(info, array);
				if (name == "unknown")
					continue; // KLUDGE!
#ifdef WIN32
				sprintf(buf, "%s %I64d %I64d\n", name.c_str(), type.m_Count, type.m_Size);
#endif
#ifdef LINUX
				sprintf(buf, "%s %qd %qd\n", name.c_str(), type.m_Count, type.m_Size);
#endif
				s += buf;
			}
			//printf("getHeapInfo: 3\n"); fflush(stdout);
		}
		//
		jbyteArray res = env->NewByteArray(s.length());
		env->SetByteArrayRegion(res, 0, s.length(), (jbyte*)s.c_str());
		//printf("getHeapInfo: 4\n"); fflush(stdout);
		//return s.length();
		return res;
	}
	JNIEXPORT jbyteArray JNICALL Java_ru_nlmk_eclipse_plugins_profiler_trace_Trace_getSamplingData(JNIEnv *env, jclass that) {
		TCSLocker csl( g_SamplingCS );
		g_OutputString.resize(0);
		try {
			TJVMInfo::instance()->print();
		}
		catch ( ... ) {
		}
		//
		/*printf("%s\n", g_OutputString.c_str());
		fflush(stdout);*/
		//
		jbyteArray result = env->NewByteArray(g_OutputString.length());
		env->SetByteArrayRegion(result, 0, g_OutputString.length(), (jbyte*)g_OutputString.c_str());
		//
		g_OutputString.resize(0);
		return result;
	}
	//JNIEXPORT jint JNICALL Java_ru_nlmk_eclipse_plugins_profiler_trace_Trace_getAllocationSites(JNIEnv *env, jclass that, jbyteArray typeNameBytes, jbyteArray res, int size) {
	JNIEXPORT jbyteArray JNICALL Java_ru_nlmk_eclipse_plugins_profiler_trace_Trace_getAllocationSites(JNIEnv *env, jclass that, jbyteArray typeNameBytes) {
		TAllocationSite sites;
		{
			TCSLocker CSL(g_CSInstances);
			//
			string typeName;
			{
				int len = env->GetArrayLength(typeNameBytes);
				char *buf = new char[len + 1];
				env->GetByteArrayRegion(typeNameBytes, 0, len, (jbyte*)buf);
				buf[len] = 0;
				typeName = buf;
			}
			//
			jobjectID typeClazz = 0;
			jint typeArray = -1;
			{
				for (TObjectToInstanceInfoMap::iterator I = g_Instances.begin(); I !=g_Instances.end(); I++) {
					jobjectID clazz = I->second.m_Class;
					jint array = I->second.m_Array;
					shared_ptr<TJVMClassInfo> info = TJVMInfo::instance()->getClass(clazz);
					string name = getTypeName(info, array);
					if (name == typeName) {
						typeClazz = clazz;
						typeArray = array;
						break;
					}
				}
			}
			if (typeArray == -1)
				return 0;
			//
			{
				for (TObjectToInstanceInfoMap::iterator I = g_Instances.begin(); I !=g_Instances.end(); I++) {
					TInstanceInfo &info = I->second;
					jobjectID clazz = info.m_Class;
					jint array = info.m_Array;
					if ((clazz != typeClazz) || (array != typeArray))
						continue;
					if ( (info.getTrace().frames == 0) || (info.getTrace().num_frames == 0) )
						continue;
					//
					sites.add(info.getTrace().frames, info.getTrace().num_frames - 0, info.m_Size);
				}
			}
		}
		//
		string s;
		sites.print(s, 0);
		//
		/*if ( s.length() > size )
			return size - s.length();*/
		jbyteArray res = env->NewByteArray(s.length());
		env->SetByteArrayRegion(res, 0, s.length(), (jbyte*)s.c_str());
		//return s.length();
		return res;
	}
}
